__author__ = """Jorge Rivera"""
__version__ = "1.3.3"

from pydeflate.deflate.deflate import deflate
from pydeflate.tools.exchange import exchange
from pydeflate.tools.update_data import update_all_data, warn_updates
from pydeflate.get_data.oecd_data import update_dac1
from pydeflate import get_data


def set_pydeflate_path(path):
    from pathlib import Path
    from pydeflate.pydeflate_config import PYDEFLATE_PATHS
    from bblocks import set_bblocks_data_path

    """Set the path to the data folder."""
    global PYDEFLATE_PATHS

    PYDEFLATE_PATHS.data = Path(path).resolve()
    set_bblocks_data_path(PYDEFLATE_PATHS.data)


# check that data is fresh enough
warn_updates()


__all__ = ["deflate", "exchange", "update_all_data", "set_pydeflate_path", "get_data"]



import io
import warnings
import zipfile as zf
from dataclasses import dataclass

import pandas as pd
import requests
from bs4 import BeautifulSoup as Bs

from pydeflate.get_data.deflate_data import Data
from pydeflate.get_data.exchange_data import ExchangeOECD
from pydeflate.pydeflate_config import PYDEFLATE_PATHS, logger
from pydeflate.tools.update_data import update_update_date
from pydeflate.utils import oecd_codes

warnings.simplefilter("ignore", Warning, lineno=1013)

_BASE_URL: str = "https://stats.oecd.org/DownloadFiles.aspx?DatasetCode="
_TABLE1_URL: str = f"{_BASE_URL}TABLE1"


def _read_zip_content(request_content: bytes, file_name: str) -> pd.DataFrame:
    """Read the contents of a zip file

    Args:
        request_content: the content of the request
        file_name: the name of the file to read.

    Returns:
        A pandas dataframe with the contents of the file

    """
    # Read the zipfile
    _ = io.BytesIO(request_content)
    zip_file = zf.ZipFile(_).open(file_name)

    # Try two alternative separators to read the CSV file
    try:
        return pd.read_csv(zip_file, sep=",", encoding="ISO-8859-1", low_memory=False)

    except UnicodeDecodeError:
        return pd.read_csv(zip_file, sep="|", encoding="ISO-8859-1", low_memory=False)


def _download_bulk_file(url: str) -> bytes:
    """Get zipfile bytes from the webpage

    Args:
        url: the url to download the file from

    Returns:
        The content of the file (bytes)

    """

    # get URL
    response = requests.get(url)

    # parse html
    page = Bs(response.text, "html.parser")
    link = list(page.find_all("a"))[0].attrs["onclick"][15:-3].replace("_", "-")
    link = f"https://stats.oecd.org/FileView2.aspx?IDFile={link}"

    return requests.get(link).content


def _clean_dac1(df: pd.DataFrame) -> pd.DataFrame:
    """Clean DAC1 to keep only relevant information for deflators and exchange.

    Args:
        df: the dataframe to clean

    Returns:
        A cleaned dataframe
    """

    # Columns to keep and rename
    cols = {
        "DONOR": "donor_code",
        "AMOUNTTYPE": "type",
        "AIDTYPE": "aid",
        "Year": "year",
        "FLOWS": "flow",
        "Value": "value",
    }

    # Get only the official definition of the data
    query = (
        "(aid == 1010 & flow == 1140 & year <2018 ) | "
        "(aid == 11010 & flow == 1160 & year >=2018)"
    )

    # Clean the data
    data = (
        df.filter(cols, axis=1)
        .rename(columns=cols)
        .astype({"year": "Int32"})
        .query(query)
        .filter(["donor_code", "type", "year", "value"])
        .pivot(index=["donor_code", "year"], columns=["type"], values="value")
        .reset_index()
        .assign(
            exchange=lambda d: round(d.N / d.A, 5),
            deflator=lambda d: round(100 * d.A / d.D, 6),  # implied deflator
            iso_code=lambda d: d.donor_code.map(oecd_codes()),
            year=lambda d: pd.to_datetime(d.year, format="%Y"),
        )
        .assign(exchange=lambda d: d.exchange.fillna(1))
        .dropna(subset=["iso_code"])
        .filter(["iso_code", "year", "exchange", "deflator"], axis=1)
    )

    return data


def update_dac1() -> None:
    """Update dac1 data from OECD site and save as feather"""

    logger.info("Downloading DAC1 data, which may take a bit")

    # File name to read
    file_name = "Table1_Data.csv"

    # Download the zip bytes
    zip_bytes = _download_bulk_file(url=_TABLE1_URL)

    # Read the zip file and clean the data
    df = (
        _read_zip_content(request_content=zip_bytes, file_name=file_name)
        .pipe(_clean_dac1)
        .reset_index(drop=True)
    )

    # Save the data
    df.to_feather(PYDEFLATE_PATHS.data / "dac1.feather")

    # Update the update date
    update_update_date("OECD DAC")

    # Log
    logger.info("Successfully downloaded DAC1 data")


def _identify_base_year(df: pd.DataFrame) -> int:

    return (
        df.groupby(["year"], as_index=False)
        .value.mean(numeric_only=True)
        .round(2)
        .loc[lambda d: d.value == 100.00]
        .year.dt.year.item()
    )


def _calculate_price_deflator(deflators_df: pd.DataFrame) -> pd.DataFrame:
    return deflators_df.assign(
        value=lambda d: round(d.value_dac * d.value_exchange / 100, 6)
    ).filter(["iso_code", "year", "indicator", "value"], axis=1)


@dataclass
class OECD(Data):
    """An object to download and return the latest OECD DAC deflators data."""

    def __post_init__(self):
        self._available_methods = {"dac_deflator": "oecd_dac"}

    def update(self, **kwargs) -> None:
        update_dac1()

    def load_data(self, **kwargs) -> None:
        """Load the OECD DAC price deflators data.

        If the data is not found, it will be downloaded.
        DAC deflators are transformed into price deflators by using the
        implied exchange rate information from the OECD DAC data.

        The deflators that are loaded is therefore *not* the DAC deflator,
        but the price deflator used to produce the DAC deflators.

        """
        try:
            pd.read_feather(PYDEFLATE_PATHS.data / "dac1.feather")
        except FileNotFoundError:
            logger.info("Data not found, downloading...")
            self.update()
        finally:
            d_ = (
                pd.read_feather(PYDEFLATE_PATHS.data / "dac1.feather")
                .assign(indicator="oecd_dac")
                .rename(columns={"deflator": "value"})
            )

        # Identify base year
        base_year = _identify_base_year(d_)

        # Load exchange deflators
        exchange_deflator = ExchangeOECD().exchange_deflator(
            source_iso="USA", target_iso="USA", base_year=base_year
        )

        # Merge deflators and exchange deflators
        deflators_df = d_.merge(
            exchange_deflator,
            on=["iso_code", "year"],
            how="left",
            suffixes=("_dac", "_exchange"),
        )

        # Calculate the price deflator
        self._data = _calculate_price_deflator(deflators_df=deflators_df)



import json

import country_converter as coco
import numpy as np
import pandas as pd

from pydeflate.pydeflate_config import PYDEFLATE_PATHS


def oecd_codes() -> dict:
    with open(PYDEFLATE_PATHS.settings / "oecd_codes.json") as file:
        updates = json.load(file)

    return {int(k): v for k, v in updates.items()}


def emu() -> list:
    with open(PYDEFLATE_PATHS.settings / "emu.json") as file:
        emu = json.load(file)

    return emu


def clean_number(number):
    """Clean a number and return as float"""
    import re

    if not isinstance(number, str):
        number = str(number)

    number = re.sub(r"[^\d.]", "", number)

    if number == "":
        return np.nan

    return float(number)


def check_year_as_number(df: pd.DataFrame, date_column: str) -> (pd.DataFrame, bool):
    """Check whether the date column contains an int instead of datetime.
    This changes the column to datetime and returns a flag"""

    if pd.api.types.is_numeric_dtype(df[date_column]):
        df[date_column] = pd.to_datetime(df[date_column], format="%Y")
        year_as_number = True
    else:
        year_as_number = False

    return df, year_as_number


def to_iso3(
    df: pd.DataFrame,
    codes_col: str,
    target_col: str,
    src_classification: str | None = None,
    not_found: str | None = None,
) -> pd.DataFrame:
    """Convert a column of country codes to iso3"""

    cc = coco.CountryConverter()

    df[target_col] = cc.pandas_convert(
        df[codes_col], src=src_classification, to="ISO3", not_found=not_found
    )

    return df



from dataclasses import dataclass

import pandas as pd
from bblocks import WorldEconomicOutlook, set_bblocks_data_path

from pydeflate.get_data.deflate_data import Data
from pydeflate.pydeflate_config import PYDEFLATE_PATHS
from pydeflate.tools.update_data import update_update_date

set_bblocks_data_path(PYDEFLATE_PATHS.data)


@dataclass
class IMF(Data):
    _weo: WorldEconomicOutlook = None
    """An object to download and return the latest IMF WEO data for several indicators"""

    def __post_init__(self):
        self._weo = WorldEconomicOutlook()
        self._available_methods = {
            "Consumer price index": "PCPI",
            "Consumer price index, end of period": "PCPIE",
            "Gross domestic product, deflator": "NGDP_D",
            "gdp": "NGDP_D",
            "pcpi": "PCPI",
            "pcpie": "PCPIE",
        }

    def update(self) -> None:
        """Update the stored WEO data, using WEO package."""
        self._weo.update_data(reload_data=False, year=None, release=None)
        update_update_date(source="IMF")

    def load_data(
        self, latest_y: int | None = None, latest_r: int | None = None
    ) -> None:
        """Loading WEO as a clean dataframe

        Args:
            latest_y: passed only optional to override the behaviour to get the latest
            release year for the WEO.
            latest_r: passed only optionally to override the behaviour to get the latest
            released value (1 or 2).
        """
        # load the data for available indicators
        indicators = list(self._available_methods.values()) + ["NGDP", "NGDPD"]
        self._weo.load_data(indicators)

        # get the data into a dataframe
        self._data = self._weo.get_data(keep_metadata=False)

    def implied_exchange(self, direction: str = "lcu_usd") -> pd.DataFrame:
        """Get the implied exchange rate used by the IMF
        Args:
            direction: the direction of the exchange rate, either lcu_usd for
            local currency units per usd, or usd_lcu for the opposite.
        """

        # USD data
        usd = self.get_method("NGDPD")

        # LCU data
        lcu = self.get_method("NGDP")

        # Merge datasets
        d_ = usd.merge(lcu, on=["iso_code", "year"], suffixes=("_usd", "_lcu"))

        # Calculate implied exchange rate
        if direction == "lcu_usd":
            d_["value"] = d_["value_lcu"] / d_["value_usd"]
        elif direction == "usd_lcu":
            d_["value"] = d_["value_usd"] / d_["value_lcu"]
        else:
            raise ValueError(f"direction must be lcu_usd or usd_lcu, not {direction}")

        return d_.filter(["iso_code", "year", "value"], axis=1)



"""Defines the deflator class"""

from dataclasses import dataclass

import numpy as np
import pandas as pd

from pydeflate.get_data.deflate_data import Data
from pydeflate.get_data.exchange_data import Exchange


@dataclass
class Deflator:
    base_year: int
    exchange_obj: Exchange
    deflator_obj: Data
    deflator_method: str
    source_currency: str = "LCU"
    target_currency: str = "USA"
    to_current: bool = False
    _xe_df: pd.DataFrame | None = None
    _xe_deflator_df: pd.DataFrame | None = None
    _price_deflator_df: pd.DataFrame | None = None
    """A class to produce deflators based on the given parameters.
    
    Deflators take into account changes in prices and the evolution of currency exchange
    values. They compensate for both factors in order to produce a standard unit of 
    measurement, regardless of the year of the flow.
    
    Args:
        base_year: the year in which the deflator equals 100 (if the source and target
        currency are equal).
        Exchange: the exchange rate object.
        price: the price data oject.
        source_currency: the ISO3 code of the country which uses the source currency
        target_currency: the ISO3 code of the country which used the target currency
        _deflator_df: the resulting DataFrame containing the deflators data.
    """

    def __post_init__(self):
        # Exchange deflator DataFrame, including any currency conversion
        self._xe_deflator_df = self.exchange_obj.exchange_deflator(
            source_iso=self.source_currency,
            target_iso=self.target_currency,
            base_year=self.base_year,
        )

        # Price deflator DataFrame
        self._price_deflator_df = self.deflator_obj.get_deflator(
            base_year=self.base_year, method=self.deflator_method
        )

    def get_deflator(self) -> pd.DataFrame:
        """Get the deflator DataFrame"""

        # Merge the price and currency exchange deflators
        df = self._price_deflator_df.merge(
            self._xe_deflator_df,
            on=["iso_code", "year"],
            how="left",
            suffixes=("_def", "_xe"),
        )

        # Calculate the deflator
        df = (
            df.assign(
                value=round(100 * df.value_def / df.value_xe, 6)
                if not self.to_current
                else round(100 * df.value_xe / df.value_def, 6)
            )
            .filter(["iso_code", "year", "value"], axis=1)
            .assign(value=lambda d: d.value.replace(0, np.nan))
            .dropna(subset=["value"])
            .reset_index(drop=True)
            .rename(columns={"value": "deflator"})
        )

        return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydeflate', 'pydeflate.deflate', 'pydeflate.get_data', 'pydeflate.tools']

package_data = \
{'': ['*'], 'pydeflate': ['.pydeflate_data/*', 'settings/*']}

install_requires = \
['bblocks>=1.0.2,<2.0.0',
 'beautifulsoup4>=4.11.2,<5.0.0',
 'country-converter>=0.8.0,<0.9.0',
 'numpy>=1.24.1,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'pyarrow>=11.0.0,<12.0.0',
 'requests>=2.28.2,<3.0.0',
 'weo>=0.7.4,<0.8.0',
 'xlrd>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'pydeflate',
    'version': '1.3.3',
    'description': 'Package to convert current prices figures to constant prices and vice versa',
    'long_description': '# pydeflate\n\n[![pypi](https://img.shields.io/pypi/v/pydeflate.svg)](https://pypi.python.org/pypi/pydeflate)\n[![Documentation Status](https://readthedocs.org/projects/pydeflate/badge/?version=latest)](https://pydeflate.readthedocs.io/en/latest/?version=latest)\n[![black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Downloads](https://pepy.tech/badge/pydeflate/month)](https://pepy.tech/project/pydeflate)\n[![Coverage](https://codecov.io/gh/jm-rivera/pydeflate/branch/main/graph/badge.svg?token=uwKI5DyO3w)](https://codecov.io/gh/jm-rivera/pydeflate)\n\n**pydeflate** is a Python package to convert flows data to constant\nprices. This can be done from any source currency to any desired base\nyear and currency. **Pydeflate** can also be used to convert constant\ndata to current prices and to convert from one currency to another (in\ncurrent and constant prices). Users can choose the source of the\nexchange and deflator/prices data (currently, IMF, World Bank or OECD\nDAC).\n\n## Getting started\n\n### Installation\n\npydeflate can be installed from PyPI. From the command line:\n\n```bash\n\npip install pydeflate\n\n```\n\n## Basic usage\n\n### Current to constant\n\nConvert data expressed in current USD prices to constant EUR prices for\na given base year:\n\n```python\nfrom pydeflate import deflate, set_pydeflate_path\nimport pandas as pd\n\n# Specify where the deflator and exchange data should be saved\nset_pydeflate_path("path/to/data/folder")\n\n# example data\ndata = {\n    \'iso_code\': [\'FRA\', \'USA\', \'GTM\'],\n    \'year\': [2017, 2017, 2017],\n    \'value\': [50, 100, 200]\n}\n\n# create an example dataframe, in current USD prices\ndf = pd.DataFrame.from_dict(data)\n\n# convert to EUR 2015 constant prices\ndf_constant = deflate(\n    df=df,\n    base_year=2015,\n    deflator_source=\'world_bank\',\n    deflator_method=\'gdp\',\n    exchange_source="world_bank",\n    source_currency="USA",  # since data is in USD\n    target_currency="EMU",  # we want the result in constant EUR\n    id_column="iso_code",\n    id_type="ISO3",  # specifying this is optional in most cases\n    date_column="year",\n    source_column="value", # where the original data is\n    target_col="value_constant", # where the new data will be stored\n)\n```\n\nThis results in a dataframe containing a new column `value_constant` in\n2015 constant prices. In the background, pydeflate takes into account:\n\n- changes in princes, through a gdp deflator in this case\n- changes in exchange rates overtime\n\n### Current Local Currency Units to constant in a different currency\nPydeflate can also handle data that is expressed in local currency\nunits. In that case, users can specify `LCU` as the source currency.\n\n```python\nfrom pydeflate import deflate, set_pydeflate_path\nimport pandas as pd\n\n# Specify where the deflator and exchange data should be saved\nset_pydeflate_path("path/to/data/folder")\n\n# example data\ndata = {\n    \'country\': [\'United Kingdom\', \'United Kingdom\', \'Japan\'],\n    \'date\': [2011, 2015, 2015],\n    \'value\': [100, 100, 100]\n}\n\n# create an example dataframe, in current local currency units\ndf = pd.DataFrame.from_dict(data)\n\n# convert to USD 2018 constant prices\ndf_constant = deflate(\n    df=df,\n    base_year=2018,\n    deflator_source=\'imf\',\n    deflator_method=\'pcpi\',\n    exchange_source="imf",\n    source_currency="LCU",  # since data is in LCU\n    target_currency="USA",  # to get data in USD\n    id_column="iso_code",\n    date_column="date",\n    source_col="value",\n    target_col="value",  # to not create a new column\n)\n\n```\n\n### Constant to current\nUsers can also convert a dataset expressed in constant prices to current\nprices using pydeflate. To avoid introducing errors, users should know\nwhich methodology/ data was used to create constant prices by the\noriginal source. The basic usage is the same as before, but the\n`to_current` parameter is set to `True`.\n\nFor example, to convert DAC data expressed in 2016 USD constant prices\nto current US dollars:\n\n```python\nfrom pydeflate import deflate, set_pydeflate_path\nimport pandas as pd\n\n# Specify where the deflator and exchange data should be saved\nset_pydeflate_path("path/to/data/folder")\n\n# example data\ndata = {\n    \'dac_code\': [302, 4, 4],\n    \'date\': [2010, 2016, 2018],\n    \'value\': [100, 100, 100]\n}\n\n# create an example dataframe, in current local currency units\ndf = pd.DataFrame.from_dict(data)\n\n# convert to USD 2018 constant prices\ndf_current = deflate(\n    df=df,\n    base_year=2016,\n    deflator_source=\'oecd_dac\',\n    deflator_method=\'dac_deflator\',\n    exchange_source="oecd_dac",\n    source_currency="USA",  # since data is in USD constant\n    target_currency="LCU",  # to get the current LCU figures\n    id_column="dac_code",\n    id_type="DAC",\n    date_column="date",\n    source_column="value",\n    target_column="value_current",\n    to_current=True,\n)\n```\n\n## Data source and method options\n\nIn order to convert the data, pydeflate uses data on **price/gdp deflators** and\n**exchange rates**. Each of these data sources can come from the `OECD DAC`,\n`IMF (WEO)` or `World Bank`.\n\nFor all sources, Exchange rates between two non USD currency pairs are derived from\nthe LCU to USD exchange rates selected.\n\n### International Monetary Fund World Economic Outlook ("imf")\n\nFor price/gdp deflators from the IMF, the following options are available (`deflator_method`):\n- `gdp`: in order to use GDP deflators.\n- `pcpi`: in order to use Consumer Price Index data.\n- `pcpie`: to use end-of-period Consumer Price Index data\n  (e.g. for December each year).\n\nThe IMF provides estimates where data is not available, including for several\nyears into the future. Using these price deflators, combined with the corresponding\nexchange rates, allows users to convert data to constant prices for future years.\n\nFor exchange rates, the following options are available from the imf (`exchange_method`):\n- `implied`: to use the exchange rates used by the World Economic Outlook, derived from\n    the WEOs data on GDP in US Dollars and Local Currency Units.\n\n### World Bank ("world_bank")\n\nFor price/gdp deflators from the World Bank, the following options are available (`deflator_method`):\n\n\nIn terms of price or GDP deflators, pydeflate provides the following\n- `gdp`: in order to use GDP deflators.\n- `gdp_linked`: to use the World Bank\'s GDP deflator series which\n  has been linked to produce a consistent time series to\n  counteract breaks in series over time due to changes in base\n  years, sources or methodologies.\n- `cpi`: to use Consumer Price Index data\n\nFor exchange rates, the following options are available from the World Bank (`exchange_method`):\n- `yearly_average`: as used by the World Bank, based on IMF International Financial Statistics data.\n\n\n### OECD Development Assistance Committee ("oecd_dac")\n\nFor price/gdp deflators from the OECD DAC, the following options are available (`deflator_method`):\n\nIn terms of price or GDP deflators, pydeflate provides the following:\n- `dac_deflator`: in order to use the DAC\'s own deflator series.\n\nFor exchange rates, the following options are available from the OECD DAC (`exchange_method`):\n- `implied`: to use the exchange rates used and published by the DAC.\n\n\n\n## Additional features\n\nPydeflate relies on data from the World Bank, IMF and OECD for its\ncalculations. This data is updated periodically. If the version of the\ndata stored in the user\'s computer is older than 50 days, pydeflate will\nshow a warning on import.\n\nUsers can always update the underlying data by using:\n\n```python\nimport pydeflate\n\npydeflate.update_all_data()\n\n```\n\nPydeflate also provides users with a tool to exchange figures from one\ncurrency to another, without applying any deflators. This should only be\nused on numbers expressed in current prices, however.\n\nFor example, to convert numbers in current Local Currency Units (LCU) to\ncurrent Canadian Dollars:\n\n```python\nimport pydeflate\nimport pandas as pd\n\n# example data\ndata = {\n    \'iso_code\': [\'GBR\', \'CAN\', \'JPN\'],\n    \'date\': [2011, 2015, 2015],\n    \'value\': [100, 100, 100]\n}\n\n# create an example dataframe, in current local currency units\ndf = pd.DataFrame.from_dict(data)\n\n# convert to USD 2018 constant prices\ndf_can = pydeflate.exchange(\n    df=df,\n    source_currency="LCU",  # since data is in LCU\n    target_currency="CAN",  # to get data in Canadian Dollars\n    rates_source=\'imf\', \n    value_column=\'value\',\n    target_column=\'value_CAN\',\n    id_column="iso_code",\n    id_type="ISO3",\n    date_column="date",\n)\n\n```\n\n### Credits\n\nThis package relies on data from the following sources:\n\n- OECD DAC: <https://www.oecd.org/dac/> (Official Development\n  assistance data (DAC1), DAC deflators, and exchange rates used by\n  the DAC)\n- IMF World Economic Outlook:\n  <https://www.imf.org/en/Publications/WEO> (GDP and price deflators)\n- World Bank DataBank: <https://databank.worldbank.org/home.aspx>\n  (exchange rates, GDP and price deflators)\n\nThis data is provided based on the terms and conditions set by the\noriginal sources.',
    'author': 'Jorge Rivera',
    'author_email': 'jorge.rivera@one.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass

import pandas as pd
from pydeflate.get_data.deflate_data import Data
from pydeflate.pydeflate_config import PYDEFLATE_PATHS

from bblocks import WorldBankData, set_bblocks_data_path

from pydeflate.tools.update_data import update_update_date

set_bblocks_data_path(PYDEFLATE_PATHS.data)

_INDICATORS: dict = {
    "gdp": "NY.GDP.DEFL.ZS",
    "gdp_linked": "NY.GDP.DEFL.ZS.AD",
    "cpi": "FP.CPI.TOTL",
    "exchange": "PA.NUS.FCRF",
    "effective_exchange": "PX.REX.REER",
}

START: int = 1950
END: int = 2025


def update_world_bank_data() -> None:
    """Update World Bank data."""
    wb = WorldBankData()
    wb.load_data(indicator=list(_INDICATORS.values()), start_year=START, end_year=END)
    wb.update_data()
    update_update_date(source="World Bank")


@dataclass
class WorldBank(Data):
    """An object to download and return the latest WorldBank exchange and price data"""

    def __post_init__(self):
        self._available_methods = {
            "gdp": "GDP deflator",
            "gdp_linked": "GDP deflator linked",
            "cpi": "Consumer price index",
        }

    def update(self, **kwargs) -> None:
        """Update data for all WorldBank indicators"""
        update_world_bank_data()

    def load_data(self) -> None:
        """Load data for all WorldBank indicators"""

        # get the paths to the data
        paths = [
            PYDEFLATE_PATHS.data / f"{_INDICATORS[i_]}_{START}-{END}_.csv"
            for i_ in _INDICATORS
        ]

        # check if data exists, if not update
        for path in paths:
            if not path.exists():
                update_world_bank_data()
                break

        # load the data
        files = []
        for path in paths:
            try:
                files.append(
                    pd.read_csv(path, parse_dates=["date"]).rename(
                        columns={"date": "year"}
                    )
                )
            except FileNotFoundError:
                files.append(pd.DataFrame())

        self._data = pd.concat(files, ignore_index=True)



import pandas as pd

from pydeflate.get_data.exchange_data import (
    ExchangeIMF,
    ExchangeOECD,
    ExchangeWorldBank,
)
from pydeflate.utils import check_year_as_number, to_iso3, oecd_codes


_exchange_source = {
    "world_bank": ExchangeWorldBank,
    "wb": ExchangeWorldBank,
    "oecd_dac": ExchangeOECD,
    "imf": ExchangeIMF,
}


def _check_key_errors(
    rates_source: str,
    columns: str | list | pd.Index,
    value_column: str,
    date_column: str,
) -> None:
    """Check whether provided parameters are valid"""

    if rates_source not in _exchange_source.keys():
        raise KeyError(
            f"{rates_source=} is not a valid exchange rates source. "
            f"Please choose from {_exchange_source.keys()}"
        )

    if value_column not in columns:
        raise KeyError(
            f"{value_column} is not a valid column in the provided DataFrame"
        )

    if date_column not in columns:
        raise KeyError(f"{date_column} is not a valid column in the provided DataFrame")


def exchange(
    df: pd.DataFrame,
    source_currency: str,
    target_currency: str,
    rates_source: str = "world_bank",
    id_column: str = "iso_code",
    id_type: str = "ISO3",
    value_column: str = "value",
    target_column: str = "value",
    date_column: str = "date",
) -> pd.DataFrame:
    """


    Parameters
    ----------
    df : pd.DataFrame
        A Pandas DataFrame, in long format, containing at least a date column,
        a column with iso-3 codes to identify the source currency, and a
        value column where the values to be converted are stored.
    source_currency : str
        The ISO-3 code of the country which owns the currency in which the data
        is expressed. "LCU" can be used to indicate that data is in Local
        Currency Unit. "emu" can be used for the EURO.
    target_currency : str
        The ISO-3 code of the country which owns the currency to which the data
        will be converted. "LCU" can be used to convert from a given currency
        (like the USD), back to each country's Local Currency.
    rates_source : str, optional
        The source of the exchange rate data. Current options include "wb" for
        the World Bank and "oecd_dac" for the exchange rates used for ODA
        statistics. The default is "wb".
    id_column : str, optional
        The name of the column containing the codes or names used to identify countries.
        The default is "iso_code".
    id_type : str, optional
        The types of codes used to identify countries. Should match options in
        Country Converter or the DAC codes.The default is "ISO3".
    value_column : str, optional
        The name of the column containing the values to be converted.
        The default is "value".
    target_column : str, optional
        The name of the column where the converted values will be stored.
        The default is "value_xe".
    date_column : str, optional
        The name of the column where the date/year is stored.
        The default is "date".

    Returns
    -------
    df : pd.DataFrame
        Returns a dataframe containing the converted data stored in the
        target column.

    """

    # create a copy of the dataframe to avoid modifying the original
    df = df.copy(deep=True)

    # Check whether provided parameters are valid
    _check_key_errors(rates_source, df.columns, value_column, date_column)

    # If source currency matches target currency, do nothing
    if source_currency == target_currency:
        df[target_column] = df[value_column]
        return df

    # keep track of original columns. This is so that the same order and columns can be
    # preserved.
    if target_column not in df.columns:
        cols = [*df.columns, target_column]
    else:
        cols = df.columns

    # check whether date is provided as integer
    df, year_as_number = check_year_as_number(df, date_column)

    # check whether target currency is LCU
    if target_currency == "LCU":
        target_currency = source_currency
        source_currency = "LCU"
        target_changed = True
    else:
        target_changed = False

    # get the selected rates function
    exchange_rates = (
        _exchange_source[rates_source]()
        .exchange_rate(target_currency)
        .rename(columns={"year": date_column, "value": value_column, "iso_code": "id_"})
    )

    # Create ID col.
    if id_type == "DAC":
        df["id_"] = df[id_column].map(oecd_codes()).fillna("DAC")
    else:
        df = df.pipe(
            to_iso3, codes_col=id_column, target_col="id_", src_classification=id_type
        )

    # merge exchange rates with data
    if source_currency == "LCU":
        df = df.merge(
            exchange_rates,
            on=["id_", date_column],
            suffixes=("", "_xe"),
        )

    else:
        xe = exchange_rates.loc[exchange_rates.id_ == source_currency]
        df = df.merge(
            xe.drop("id_", axis=1),
            on=[date_column],
            suffixes=("", "_xe"),
        )

    # revert change to target_currency if target_changed
    if target_changed:
        target_currency = "LCU"

    if target_currency == "LCU":
        df[target_column] = df[value_column] * df[f"{value_column}_xe"]

    else:
        df[target_column] = df[value_column] / df[f"{value_column}_xe"]

    if year_as_number:
        df[date_column] = df[date_column].dt.year

    return df.filter(cols, axis=1)
